from __future__ import annotations

import json
import socket
import threading
import time
from urllib.error import HTTPError
from urllib.request import Request as UrlRequest, urlopen

from orchesis.demo import DEMO_STATS, DemoServer
from tests.cli_test_utils import CliRunner
from orchesis.cli import main


def _pick_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = int(sock.getsockname()[1])
    sock.close()
    return port


def _get_json(url: str) -> dict:
    with _urlopen_with_retry(url, timeout=5) as resp:
        return json.loads(resp.read().decode("utf-8"))


def _urlopen_with_retry(request_or_url, timeout: int = 5, attempts: int = 3):
    last_error = None
    for _ in range(max(1, attempts)):
        try:
            return urlopen(request_or_url, timeout=timeout)
        except Exception as error:
            last_error = error
            time.sleep(0.05)
    if last_error is not None:
        raise last_error
    return urlopen(request_or_url, timeout=timeout)


def _wait_for_server(port: int, attempts: int = 20) -> None:
    for _ in range(max(1, attempts)):
        try:
            with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=1, attempts=1):
                return
        except Exception:
            time.sleep(0.05)
    raise RuntimeError(f"demo server did not start on port {port}")


def test_demo_stats_valid_json() -> None:
    assert isinstance(DEMO_STATS, dict)
    assert "cost_velocity" in DEMO_STATS


def test_demo_server_starts() -> None:
    port = _pick_port()
    thread = threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True)
    thread.start()
    _wait_for_server(port)
    with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=5) as resp:
        assert int(resp.status) == 200


def test_demo_dashboard_has_banner() -> None:
    port = _pick_port()
    thread = threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True)
    thread.start()
    _wait_for_server(port)
    with _urlopen_with_retry(f"http://127.0.0.1:{port}/dashboard", timeout=5) as resp:
        body = resp.read().decode("utf-8")
    assert "DEMO MODE" in body


def test_demo_stats_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/stats")
    assert "cost_today" in payload
    assert "cost_velocity" in payload


def test_demo_agents_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/dashboard/agents")
    assert "agents" in payload
    assert isinstance(payload["agents"], list)


def test_demo_mode_no_real_proxy() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(
        f"http://127.0.0.1:{port}/v1/chat/completions",
        data=b"{}",
        method="POST",
    )
    try:
        _urlopen_with_retry(req, timeout=5)
        assert False
    except HTTPError as error:
        assert error.code == 404


def test_demo_approvals_endpoint() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/v1/approvals")
    assert "pending" in payload
    assert "history" in payload


def test_demo_approve_action() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(f"http://127.0.0.1:{port}/api/v1/approvals/demo-approval-1/approve", data=b"{}", method="POST")
    with _urlopen_with_retry(req, timeout=5) as resp:
        payload = json.loads(resp.read().decode("utf-8"))
    assert payload["approved"] is True


def test_demo_deny_action() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    req = UrlRequest(f"http://127.0.0.1:{port}/api/v1/approvals/demo-approval-2/deny", data=b"{}", method="POST")
    with _urlopen_with_retry(req, timeout=5) as resp:
        payload = json.loads(resp.read().decode("utf-8"))
    assert payload["denied"] is True


def test_demo_overview_contains_compliance() -> None:
    port = _pick_port()
    threading.Thread(target=DemoServer().start, kwargs={"port": port}, daemon=True).start()
    _wait_for_server(port)
    payload = _get_json(f"http://127.0.0.1:{port}/api/dashboard/overview")
    assert "compliance_overview" in payload


def test_demo_starts_both_services(monkeypatch) -> None:
    import orchesis.cli as cli_module
    import orchesis.proxy as proxy_module

    events: list[str] = []

    class _FakeProxy:
        def __init__(self, policy_path=None, config=None):
            _ = (policy_path, config)
            events.append("proxy_init")

        def start(self, blocking: bool = False) -> None:
            events.append(f"proxy_start:{bool(blocking)}")

        def stop(self) -> None:
            events.append("proxy_stop")

    class _FakeHTTPProxyConfig:
        def __init__(self, host: str = "127.0.0.1", port: int = 8080):
            _ = host
            _ = port

    class _FakeUvicorn:
        class Config:
            def __init__(self, app, host: str, port: int, log_level: str):
                _ = (app, host, port, log_level)

        class Server:
            def __init__(self, cfg):
                _ = cfg
                self.should_exit = False

            def run(self) -> None:
                events.append("api_run")

    def _fake_create_api_app(policy_path: str = "policy.yaml"):
        events.append(f"api_app:{policy_path}")
        return object()

    monkeypatch.setattr(proxy_module, "LLMHTTPProxy", _FakeProxy)
    monkeypatch.setattr(proxy_module, "HTTPProxyConfig", _FakeHTTPProxyConfig)
    monkeypatch.setattr(
        cli_module,
        "_load_server_runtime",
        lambda: (_FakeUvicorn, _fake_create_api_app, object, object),
    )

    def _stop_loop(_seconds: float) -> None:
        raise KeyboardInterrupt

    monkeypatch.setattr(cli_module.time, "sleep", _stop_loop)

    runner = CliRunner()
    result = runner.invoke(main, ["demo", "--port", "8080", "--api-port", "8090"])
    assert result.exit_code == 0
    assert "Orchesis proxy:    http://localhost:8080" in result.output
    assert "Orchesis API:      http://localhost:8090" in result.output
    assert "Dashboard:         http://localhost:8080/dashboard" in result.output
    assert "Overwatch API:     http://localhost:8090/api/v1/overwatch" in result.output
    assert "proxy_start:False" in events
    assert "api_run" in events




"""Context profile JSONL logger for per-request metrics."""

from __future__ import annotations

import json
import time
from pathlib import Path
from typing import Iterable


DEFAULT_LOG_PATH = Path(".orchesis/context_profile.jsonl")


def infer_task_type(tool_calls: Iterable[str]) -> str:
    """Infer task type from tool names."""
    names = [str(name).strip().lower() for name in tool_calls if isinstance(name, str)]
    if not names:
        return "unknown"

    coding_tokens = (
        "code",
        "python",
        "js",
        "typescript",
        "compile",
        "test",
        "build",
        "git",
        "patch",
        "refactor",
    )
    research_tokens = (
        "search",
        "web",
        "browse",
        "fetch",
        "read",
        "query",
        "lookup",
        "wikipedia",
    )
    planning_tokens = (
        "plan",
        "roadmap",
        "todo",
        "task",
        "strategy",
        "design",
        "spec",
    )

    for name in names:
        if any(token in name for token in coding_tokens):
            return "coding"
    for name in names:
        if any(token in name for token in research_tokens):
            return "research"
    for name in names:
        if any(token in name for token in planning_tokens):
            return "planning"
    return "unknown"


def log_context_profile(
    *,
    session_id: str,
    agent_id: str,
    prompt_length: int,
    tool_calls: list[str] | tuple[str, ...] | None,
    model: str,
    latency_ms: float,
    retry_count: int,
    compression_ratio: float = 1.0,
    log_path: str | Path = DEFAULT_LOG_PATH,
) -> None:
    """
    Append one request profile to JSONL.

    Safe and non-blocking: all I/O errors are swallowed intentionally.
    """
    tools = [str(item) for item in (tool_calls or []) if isinstance(item, str)]
    payload = {
        "timestamp": float(time.time()),
        "session_id": str(session_id or "unknown"),
        "agent_id": str(agent_id or "unknown"),
        "prompt_length": max(0, int(prompt_length or 0)),
        "tool_calls": tools,
        "model": str(model or ""),
        "latency_ms": float(latency_ms or 0.0),
        "retry_count": max(0, int(retry_count or 0)),
        "task_type": infer_task_type(tools),
        "compression_ratio": float(compression_ratio or 1.0),
    }

    try:
        out_path = Path(log_path)
        out_path.parent.mkdir(parents=True, exist_ok=True)
        with out_path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps(payload, ensure_ascii=False) + "\n")
    except Exception:
        # Must never block or fail the main request path.
        return



from __future__ import annotations

from tests.cli_test_utils import CliRunner

from orchesis.cli import main


class _FakeProcess:
    def __init__(self, command: list[str], return_code: int = 0) -> None:
        self.command = list(command)
        self.return_code = int(return_code)
        self.terminated = False
        self.killed = False
        self.wait_calls = 0

    def wait(self, timeout: float | None = None) -> int:
        _ = timeout
        self.wait_calls += 1
        return self.return_code

    def terminate(self) -> None:
        self.terminated = True

    def kill(self) -> None:
        self.killed = True


def test_launch_starts_proxy(monkeypatch) -> None:  # noqa: ANN001
    commands: list[list[str]] = []

    def _fake_popen(cmd, *args, **kwargs):  # noqa: ANN001
        _ = args, kwargs
        commands.append(list(cmd))
        return _FakeProcess(list(cmd))

    monkeypatch.setattr("orchesis.cli.subprocess.Popen", _fake_popen)
    monkeypatch.setattr("orchesis.cli.time.sleep", lambda *_args, **_kwargs: None)
    runner = CliRunner()
    result = runner.invoke(main, ["launch", "openclaw"])
    assert result.exit_code == 0
    assert len(commands) >= 1
    assert commands[0][:4] == ["python", "-m", "orchesis", "proxy"] or commands[0][1:4] == [
        "-m",
        "orchesis",
        "proxy",
    ]


def test_launch_connects_agent(monkeypatch) -> None:  # noqa: ANN001
    commands: list[list[str]] = []

    def _fake_popen(cmd, *args, **kwargs):  # noqa: ANN001
        _ = args, kwargs
        commands.append(list(cmd))
        return _FakeProcess(list(cmd))

    monkeypatch.setattr("orchesis.cli.subprocess.Popen", _fake_popen)
    monkeypatch.setattr("orchesis.cli.time.sleep", lambda *_args, **_kwargs: None)
    runner = CliRunner()
    result = runner.invoke(main, ["launch", "aider"])
    assert result.exit_code == 0
    assert len(commands) >= 2
    assert commands[1] == ["aider", "--openai-api-base", "http://localhost:8080/v1"]


def test_launch_cleanup_on_exit(monkeypatch) -> None:  # noqa: ANN001
    created: list[_FakeProcess] = []

    def _fake_popen(cmd, *args, **kwargs):  # noqa: ANN001
        _ = args, kwargs
        process = _FakeProcess(list(cmd))
        created.append(process)
        return process

    monkeypatch.setattr("orchesis.cli.subprocess.Popen", _fake_popen)
    monkeypatch.setattr("orchesis.cli.time.sleep", lambda *_args, **_kwargs: None)
    runner = CliRunner()
    result = runner.invoke(main, ["launch", "codex"])
    assert result.exit_code == 0
    assert len(created) >= 2
    proxy_process = created[0]
    assert proxy_process.terminated is True


def test_launch_invalid_agent() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["launch", "unknown-agent"])
    assert result.exit_code != 0
    assert "not one of" in result.output.lower() or "invalid choice" in result.output.lower()



"""Shared request-pipeline helpers."""

from __future__ import annotations

from typing import Any

from orchesis.agent_store import AgentPolicyStore


def check_budget(
    agent_id: str,
    *,
    policy_store: AgentPolicyStore,
    decisions_log_path: str,
) -> tuple[bool, dict[str, Any]]:
    """
    Pre-request check.
    Blocks if spent today is already at/above agent budget.
    """
    policy = policy_store.get_policy(agent_id)
    limit = policy.get("budget_daily")
    if not isinstance(limit, int | float):
        return True, {}
    spent = policy_store.get_cost_today(agent_id, decisions_log_path)
    safe_limit = float(limit)
    if spent >= safe_limit:
        return (
            False,
            {
                "blocked": True,
                "reason": "budget_exceeded",
                "spent": spent,
                "limit": safe_limit,
                "message": f"Daily budget ${safe_limit:.2f} reached (spent: ${spent:.2f})",
            },
        )
    return True, {}



"""Append-only evidence ledger with SHA-256 hash chaining."""

from __future__ import annotations

import hashlib
import json
import time
from pathlib import Path
from typing import Any


class EvidenceLedger:
    """Store audit evidence entries as a tamper-evident hash chain."""

    def __init__(self, path: str | Path = ".orchesis/evidence_ledger.jsonl") -> None:
        self._path = Path(path)
        self._last_hash = self._load_last_hash()

    @staticmethod
    def _hash_payload(event: dict[str, Any], timestamp: float, prev_hash: str) -> str:
        material = {
            "event": event,
            "timestamp": float(timestamp),
            "prev_hash": str(prev_hash),
        }
        encoded = json.dumps(material, ensure_ascii=False, sort_keys=True, separators=(",", ":")).encode("utf-8")
        return hashlib.sha256(encoded).hexdigest()

    def _load_last_hash(self) -> str:
        if not self._path.exists():
            return ""
        last_valid_hash = ""
        try:
            with self._path.open("r", encoding="utf-8") as handle:
                for line in handle:
                    line = line.strip()
                    if not line:
                        continue
                    try:
                        row = json.loads(line)
                    except Exception:
                        continue
                    if isinstance(row, dict):
                        hash_value = row.get("hash")
                        if isinstance(hash_value, str):
                            last_valid_hash = hash_value
        except Exception:
            return ""
        return last_valid_hash

    def record(self, event: dict[str, Any]) -> str:
        """Append one event to the ledger and return its hash."""
        if not isinstance(event, dict):
            raise TypeError("event must be a dict")
        timestamp = float(time.time())
        prev_hash = self._last_hash
        hash_value = self._hash_payload(event, timestamp, prev_hash)
        entry = {
            "event": event,
            "timestamp": timestamp,
            "prev_hash": prev_hash,
            "hash": hash_value,
        }
        self._path.parent.mkdir(parents=True, exist_ok=True)
        with self._path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps(entry, ensure_ascii=False) + "\n")
        self._last_hash = hash_value
        return hash_value

    def verify_chain(self) -> bool:
        """Verify hash chain integrity for all ledger entries."""
        if not self._path.exists():
            return True

        expected_prev = ""
        try:
            with self._path.open("r", encoding="utf-8") as handle:
                for line in handle:
                    line = line.strip()
                    if not line:
                        continue
                    row = json.loads(line)
                    if not isinstance(row, dict):
                        return False

                    event = row.get("event")
                    timestamp = row.get("timestamp")
                    prev_hash = row.get("prev_hash")
                    hash_value = row.get("hash")

                    if not isinstance(event, dict):
                        return False
                    if not isinstance(timestamp, int | float):
                        return False
                    if not isinstance(prev_hash, str) or not isinstance(hash_value, str):
                        return False
                    if prev_hash != expected_prev:
                        return False

                    computed = self._hash_payload(event, float(timestamp), prev_hash)
                    if computed != hash_value:
                        return False

                    expected_prev = hash_value
        except Exception:
            return False
        return True
